@@ -31,6 +31,7 @@
 import socket
 from typing import Optional
 
+from startrek.types import Address
 from startrek.net.channel import is_opened
 from startrek import BaseChannel, ChannelReader, ChannelWriter
 
@@ -38,7 +39,7 @@ from startrek import BaseChannel, ChannelReader, ChannelWriter
 class StreamChannelReader(ChannelReader):
 
     # Override
-    def receive(self, max_len: int) -> (Optional[bytes], Optional[tuple]):
+    def receive(self, max_len: int) -> (Optional[bytes], Optional[Address]):
         data = self.read(max_len=max_len)
         if data is None or len(data) == 0:
             return None, None
@@ -49,7 +50,7 @@ class StreamChannelReader(ChannelReader):
 class StreamChannelWriter(ChannelWriter):
 
     # Override
-    def send(self, data: bytes, target: tuple) -> int:
+    def send(self, data: bytes, target: Address) -> int:
         # TCP channel will be always connected
         # so the target address must be the remote address
         remote = self.remote_address
@@ -60,21 +61,25 @@ class StreamChannelWriter(ChannelWriter):
 class StreamChannel(BaseChannel):
     """ Stream Channel """
 
-    def __init__(self, remote: Optional[tuple], local: Optional[tuple], sock: socket.socket):
-        super().__init__(remote=remote, local=local, sock=sock)
+    def __init__(self, remote: Optional[Address], local: Optional[Address], sock: socket.socket):
+        super().__init__(remote=remote, local=local)
         self.__sock = sock
+        self._refresh_flags(sock=sock)
+
+    # Override
+    def _get_socket(self) -> Optional[socket.socket]:
+        return self.__sock
 
     # Override
     def _set_socket(self, sock: Optional[socket.socket]):
-        super()._set_socket(sock=sock)
+        # 1. check old socket
+        old = self._get_socket()
+        if old is not None and old is not sock:
+            if is_opened(sock=old):
+                close_socket(sock=old)
+        # 2. set new socket
         self.__sock = sock
 
-    # Override
-    def _close_socket(self, sock: socket.socket):
-        if is_opened(sock=sock):
-            # sock.shutdown(socket.SHUT_RDWR)
-            sock.close()
-
     # Override
     def _create_reader(self):
         return StreamChannelReader(channel=self)
@@ -82,3 +87,11 @@ class StreamChannel(BaseChannel):
     # Override
     def _create_writer(self):
         return StreamChannelWriter(channel=self)
+
+
+def close_socket(sock: socket.socket):
+    try:
+        # sock.shutdown(socket.SHUT_RDWR)
+        sock.close()
+    except socket.error:
+        pass
